import requests

"""
Default values to fall back on for required fields.
"""
DEFAULTS = {
    "title": "Form",
    "submit_label": "Submit",
    "element": {"label": "Field", "type": "text", "options": []},
    "ui:order": 999999,
}

"""
Input types that are actually subtypes of "text".
"""
TEXT_INPUTS = ["email", "number", "tel", "url"]

"""
Special Slack data sources for dynamic lists.
"""
SELECT_EXTERNAL_SOURCES = ["users", "channels", "external", "conversations"]


def schema_to_form(name, json, ui_schema, data={}):
    """
    Generate a Slack form dictionary from a schema and data.
    """
    ui = ui_schema if ui_schema is not None else {}

    form = {
        # title --> title (or default)
        "title": json.get("title", DEFAULTS["title"]),
        # name of form --> callback_id (or default)
        "callback_id": name,
        # ui.submit.value --> submit_label (or default)
        "submit_label": ui.get("submit", {}).get(
            "ui:value", DEFAULTS["submit_label"]
        ),
        # to be filled with prop_to_input
        "elements": [],
    }

    element_count = 0
    for prop in json.get("properties", {}).keys():
        # Limit elements to five, truncate after
        if element_count == 5:
            break
        element_count += 1

        form["elements"].append(
            prop_to_input(
                prop,
                json["properties"][prop],
                ui.get(prop, {}),
                data.get(prop, None),
                json.get("required", []),
            )
        )

    form["elements"] = sorted(
        form["elements"],
        key=lambda k: k["order"] if "order" in k else DEFAULTS["ui:order"]
    )

    return form


def prop_to_input(name, json, ui, data, required):
    """
    Generate a Slack form element dictionary from a schema and data.
    """
    input = {}

    # key of prop --> name
    input["name"] = name

    # order --> order (if exists)
    if "ui:order" in ui:
        input["order"] = ui["ui:order"]

    # title --> label (or default)
    input["label"] = json.get("title", DEFAULTS["element"]["label"])

    # widget --> type (or default)
    input["type"] = ui.get("ui:widget", DEFAULTS["element"]["type"])

    # If widget is a subtype of "text"...
    if input["type"] in TEXT_INPUTS:
        # "text" --> type
        input["type"] = "text"
        # widget --> subtype
        input["subtype"] = ui["ui:widget"]

    # placeholder --> placeholder (if exists)
    if "ui:placeholder" in ui:
        input["placeholder"] = ui["ui:placeholder"]

    # help --> hint (if exists)
    if "ui:help" in ui:
        input["hint"] = ui["ui:help"]

    # data --> value (if exists)
    if data:
        input["value"] = data
    # value --> value (if exists and no data)
    elif "ui:value" in ui:
        input["value"] = ui["ui:value"]

    # If not required, it's optional
    input["optional"] = name not in required

    # Char length validation on applies to text fields in text inputs
    isTextField = json.get("type", "") == "string"
    isTextInput = input["type"] == "text" or input["type"] == "textarea"
    if isTextField and isTextInput:
        # maxLength --> max_length (if exists)
        if "maxLength" in json:
            input["max_length"] = json["maxLength"]

        # minLength --> min_length (if exists)
        if "minLength" in json:
            input["min_length"] = json["minLength"]

    # Creating options array for select inputs
    if input["type"] == "select":
        if json.get("type", "") == "boolean":
            bools = ["True", "False"]
            names = json["enumNames"] if "enumNames" in json else bools
            input["options"] = [
                {"value": val, "label": names[idx]}
                for idx, val in enumerate(bools)
            ]
        elif "source" in json:
            if json["source"] in SELECT_EXTERNAL_SOURCES:
                # source --> data_source (if a Slack source keyword)
                input["data_source"] = json["source"]
            else:
                # options --> external feed (if URL provided as source)
                r = requests.get(url=json["source"])
                input["options"] = r.json()

        elif "enum" in json:
            # Default labels to value names if no "enunNames"
            names = json["enumNames"] if "enumNames" in json else json["enum"]

            # Create options array out of explicity stated options
            input["options"] = [
                {"value": val, "label": names[idx]}
                for idx, val in enumerate(json["enum"])
            ]
        else:
            # If no "source" or "enum" use default choices
            input["options"] = DEFAULTS["element"]["options"]

    # Handle data overrides...
    if data:
        if isTextInput:
            input["value"] = data
        elif input["type"] == "select":
            pass

    return input



# Generated by Django 2.1.4 on 2018-12-13 19:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0008_auto_20181206_1640'),
    ]

    operations = [
        migrations.CreateModel(
            name='Endpoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(help_text='The name of your endpoint.', max_length=100, unique=True)),
                ('url', models.URLField(blank=True, help_text='A URL to hit with the processed and validated form data.')),
                ('token', models.SlugField(editable=False, max_length=30, unique=True)),
            ],
        ),
        migrations.RemoveField(
            model_name='form',
            name='webhook',
        ),
        migrations.AddField(
            model_name='form',
            name='endpoint',
            field=models.ForeignKey(help_text='A URL to hit with the processed and validated form data.', null=True, on_delete=django.db.models.deletion.PROTECT, to='slackforms.Endpoint'),
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-04 17:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0005_auto_20181204_1731'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='webhook',
            field=models.URLField(blank=True, help_text='A URL to hit with the processed and validated form data.'),
        ),
    ]



import uuid

from django.db import models


class Token(models.Model):
    """
    A registered token for endpoint validation.
    """

    name = models.SlugField(
        max_length=100, unique=True, help_text="The name of your token."
    )
    token = models.SlugField(max_length=30, unique=True, editable=False)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.token:
            self.token = uuid.uuid4().hex[:30]
        super().save(*args, **kwargs)



import json
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, JsonResponse
from django.views import View
from ..models import Token
from ..conf import settings
from ..utils import slack


class Callback(View):
    """
    Endpoint that posts messages in Slack channels. Designed to provide
    in-channel feedback after a successful form process. Returns the response
    data received from Slack.
    """

    def get(self, request):
        return HttpResponse("OK", status=200)

    def post(self, request, format=None):
        # process the request data
        data = request.POST
        token = data.get("token", None)
        data_id = data.get("data_id", None)
        channel = data.get("channel", None)
        form = data.get("form", None)
        new = data.get("new", None)
        edit = data.get("edit", None)
        delete = data.get("delete", None)

        # authenticate the request
        token = data.get("token")
        if token is not None and token != settings.SLACK_VERIFICATION_TOKEN:
            if token not in [tok.token for tok in Token.objects.all()]:
                return HttpResponse("Invalid Verification Token.", status=403)

        # validate the request data
        # required fields
        for param in [("channel", channel), ("form", form)]:
            if param[1] is None:
                return HttpResponse(
                    'No "{}" provided.'.format(param[0]), status=400
                )

        # data_id is required if edit or delete buttons are being used
        if data_id is None:
            for param in [("edit", edit), ("delete", delete)]:
                if param[1] is not None:
                    return HttpResponse(
                        'No "data_id" provided with a "{}".'.format(param[0]),
                        status=400,
                    )

        # form is required if buttons are being used
        if form is None:
            for param in [("new", new), ("edit", edit), ("delete", delete)]:
                if param[1] is not None:
                    return HttpResponse(
                        'No "form" provided with a "{}".'.format(param[0]),
                        status=400,
                    )

        message = {}
        message["channel"] = channel
        message["text"] = request.POST.get("text", "")

        actions = []
        if new is not None:
            button_data = {"method": "POST", "form": form}
            button = {
                "name": "new",
                "text": new,
                "type": "button",
                "value": json.dumps(button_data),
            }
            actions.append(button)

        if edit is not None:
            button_data = {"method": "PUT", "data_id": data_id, "form": form}
            button = {
                "name": "edit",
                "text": edit,
                "type": "button",
                "value": json.dumps(button_data),
            }
            actions.append(button)

        if delete is not None:
            button_data = {
                "method": "DELETE",
                "data_id": data_id,
                "form": form,
            }
            button = {
                "name": "delete",
                "text": delete,
                "type": "button",
                "value": json.dumps(button_data),
                "style": "danger",
                "confirm": {
                    "title": "Are you sure?",
                    "text": "Do you want to delete the {} {} {}?".format(
                        form, "with an ID of", data_id
                    ),
                    "ok_text": "Yes",
                    "dismiss_text": "No",
                },
            }
            actions.append(button)
        buttons_attatchment = {
            "fallback": "",
            "callback_id": form,
            "actions": actions,
        }
        message["attachments"] = json.loads(
            request.POST.get("attachments", "[]")
        ) + [buttons_attatchment]

        resp = slack("chat.postMessage", **message)
        return JsonResponse(resp, status=200)

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



import json
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, JsonResponse
from django.views import View
from ..models import Token
from ..conf import settings
from ..utils import slack


class Message(View):
    """
    Endpoint that posts messages in Slack channels. Designed to provide
    a more flexible way to post custom messages. Returns the response
    data received from Slack.
    """

    def get(self, request):
        return HttpResponse("OK", status=200)

    def post(self, request, format=None):
        data = request.POST

        token = data.get("token")
        if token != settings.SLACK_VERIFICATION_TOKEN:
            if token not in [tok.token for tok in Token.objects.all()]:
                return HttpResponse("Invalid Verification Token.", status=403)

        message_data = json.loads(data.get("payload", ""))
        resp = slack("chat.postMessage", **message_data)
        return JsonResponse(resp, status=200)

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



# Generated by Django 2.1.4 on 2018-12-03 18:59

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Form',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField(unique=True)),
                ('webhook', models.URLField(blank=True)),
                ('json_schema', django.contrib.postgres.fields.jsonb.JSONField(null=True)),
                ('ui_schema', django.contrib.postgres.fields.jsonb.JSONField(null=True)),
            ],
        ),
    ]



import json
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse
from django.views import View

from ..conf import settings
from ..models import Token

from ..handlers import (
    FormHandler,
    ActionHandler,
    MessageHandler,
    SlashHandler,
    ManualHandler,
)

"""
Mapping reequest types to valid handler functions.
"""
HANDLERS = {
    "dialog_submission": FormHandler().handle,
    "interactive_message": MessageHandler().handle,
    "message_action": ActionHandler().handle,
    "slash_command": SlashHandler().handle,
    "manual": ManualHandler().handle,
}


class Home(View):
    """
    Endpoint to receive all interactions from Slack (i.e. slash commands,
    dialog submissions, interactive messages, message actions, and even manual
    requests sent from other servers).
    """

    def get(self, request):
        """
        Server status.
        """
        return HttpResponse("OK", status=200)

    def post(self, request, format=None):
        """
        Validate request token and pass data to the appropriate handler.
        """
        if "payload" in request.POST:
            data = json.loads(request.POST.get("payload"))
        else:
            data = request.POST

        token = data.get("token")
        if token != settings.SLACK_VERIFICATION_TOKEN:
            if token not in [tok.token for tok in Token.objects.all()]:
                return HttpResponse("Invalid Verification Token.", status=403)

        type = data.get("type", "")
        if type == "" and "command" in data:
            type = "slash_command"

        if type not in HANDLERS:
            return HttpResponse(
                "No handler for {} request".format(type), status=400
            )

        handler = HANDLERS.get(type)
        return handler(data)

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



import re
import uuid
from .models import Form, Token
from django.contrib import admin
from django import forms
from slackforms.utils.schema_to_form import (
    SELECT_EXTERNAL_SOURCES,
    TEXT_INPUTS,
)


url_pattern = (
    r"(?i)\b(?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.]"
    r"[a-z]{2,4}/)(?:[^\s()<>\|]+|\([^\s()<>\|]+|\([^\s()<>\|]+\)*\))+(?:\([^"
    r"""\s()<>\|]+|\([^\s()<>\|]+\)*\)|[^\s`!()\[\]{};:'".,<>\|?«»“”‘’])"""
)

CUSTOM_UI_SCHEMA_KEYS = ["submit"]
UI_SCHEMA_WIDGETS = ["text", "textarea", "select"] + TEXT_INPUTS


class FormForm(forms.ModelForm):
    def validate_json_schema(self):
        schema = self.cleaned_data.get("json_schema", {})
        errors = []

        if len(schema.get("required", [1, 2, 3])) == 0:
            errors.append('Remove "required" field if no props are required.')

        props = schema.get("properties", {})

        if len(props.keys()) > 5:
            errors.append("More than five properties included.")

        for prop in props:
            for attr in props[prop]:
                if attr == "source":
                    if props[prop][
                        attr
                    ] not in SELECT_EXTERNAL_SOURCES and not re.match(
                        url_pattern, props[prop][attr]
                    ):
                        errors.append(
                            "%s: source not one of %s or a valid URL."
                            % (prop, SELECT_EXTERNAL_SOURCES)
                        )
                if attr == "enumNames":
                    if len(props[prop]["enum"]) != len(
                        props[prop]["enumNames"]
                    ):
                        errors.append(
                            "%s: length of enum (%s) != enumNames(%s.)"
                            % (
                                prop,
                                props[prop]["enum"],
                                props[prop]["enumNames"],
                            )
                        )

        return None if len(errors) == 0 else errors

    def validate_ui_schema(self):
        schema = self.cleaned_data.get("ui_schema", None)
        if schema is None:
            return None

        json_schema = self.cleaned_data.get("json_schema", {})

        errors = []

        json_props = json_schema.get("properties", {}).keys()
        json_props = json_props if len(json_props) > 0 else None
        for prop in schema:
            if (
                json_props is not None
                and prop not in json_props
                and prop not in CUSTOM_UI_SCHEMA_KEYS
            ):
                errors.append(
                    "%s is not a property in the json schema." % (prop)
                )

            if schema[prop].get("ui:widget", "text") not in UI_SCHEMA_WIDGETS:
                errors.append(
                    "%s ui:widget: %s not one of %s."
                    % (prop, schema[prop].get("ui:widget"), UI_SCHEMA_WIDGETS)
                )

        return None if len(errors) == 0 else errors

    def clean(self):
        json_errors = self.validate_json_schema()
        ui_errors = self.validate_ui_schema()

        errors = {}
        if json_errors is not None:
            errors["json_schema"] = json_errors

        if ui_errors is not None:
            errors["ui_schema"] = ui_errors

        if len(errors.keys()) > 0:
            raise forms.ValidationError(errors)

        return self.cleaned_data

    class Meta:
        model = Form
        fields = [
            "name",
            "json_schema",
            "ui_schema",
            "token",
            "endpoint",
            "data_source",
            "slash_command",
        ]


class FormAdmin(admin.ModelAdmin):
    form = FormForm
    list_display = (
        "name",
        "json_schema",
        "endpoint",
    )


admin.site.register(Form, FormAdmin)


def regenerate_token(modeladmin, request, queryset):
    for token in queryset:
        token.token = uuid.uuid4().hex[:30]
        token.save()


regenerate_token.short_description = 'Regenerate endpoint token'


class TokenAdmin(admin.ModelAdmin):
    list_display = ("name", "token")
    readonly_fields = ('token',)
    actions = (regenerate_token,)


admin.site.register(Token, TokenAdmin)



# Generated by Django 2.1.4 on 2018-12-04 17:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0004_auto_20181204_1729'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='argument_prop_name',
            field=models.CharField(blank=True, help_text='The prop name of the argument of the slash command.', max_length=40, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='form',
            name='slash_command',
            field=models.CharField(blank=True, help_text="The slash command for this form (don't include the slash).", max_length=40, null=True, unique=True),
        ),
    ]



from django.http import HttpResponse, JsonResponse
from django.core.exceptions import ObjectDoesNotExist
from slackforms.models import Form


class FormHandler:
    """
    Handler for dialog submissions. Validates data based on the callback_id
    and returns errors. If no errors, it calls the form's webhook.
    """

    def handle(self, data):
        try:
            form = Form.objects.get(name=data.get("callback_id"))
        except ObjectDoesNotExist:
            return HttpResponse(
                "Form {} not found".format(data.get("callback_id")), status=404
            )

        processed = form.process(data.get("submission"))
        validation = form.validate(processed)

        if validation is True:
            meta = {
                "data_id": None
                if data.get("state") == ""
                else data.get("state"),
                "team": data.get("team"),
                "channel": data.get("channel"),
                "user": data.get("user"),
            }

            method = "POST" if meta["data_id"] is None else "PUT"
            form.post_to_webhook(processed, method, meta=meta)
            return HttpResponse(status=200)
        else:
            errors = [validation]
            return JsonResponse({"errors": errors}, safe=False)



# Generated by Django 2.1.4 on 2018-12-04 17:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0003_auto_20181204_1728'),
    ]

    operations = [
        migrations.RenameField(
            model_name='form',
            old_name='argument_name',
            new_name='argument_prop_name',
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-13 19:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0009_auto_20181213_1921'),
    ]

    operations = [
        migrations.AlterField(
            model_name='endpoint',
            name='url',
            field=models.URLField(help_text='A URL to hit with the processed and validated form data.'),
        ),
    ]



from django.urls import path

from .views import Home, Callback

urlpatterns = [
    path("", Home.as_view(), name="slackforms-home"),
    path("callback/", Callback.as_view(), name="slackforms-callback"),
]



from slackclient import SlackClient
from slackforms.conf import settings


"""
Slack API methods commonly handled by bots rather than apps.
"""
BOT_METHODS = ["chat.postMessage", "chat.postEphemeral"]


slack_api = SlackClient(settings.SLACK_API_TOKEN)
slack_bot = SlackClient(settings.SLACK_BOT_TOKEN)


def slack(method, *args, **kwargs):
    """
    Pass through to handle the different Slack tokens based on the API method
    being used.
    """
    if method in BOT_METHODS:
        return slack_bot.api_call(method, *args, **kwargs)
    else:
        return slack_api.api_call(method, *args, **kwargs)



# Generated by Django 2.1.4 on 2018-12-06 16:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0007_auto_20181205_1625'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='data_source',
            field=models.URLField(blank=True, default='http://example.com', help_text='\nThe source of data if an Id is supplied. Variable is passed through to\nthis template. Use "{id}" to crate the template.\n'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-13 19:54

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0011_auto_20181213_1953'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='json_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(help_text='A schema for the form data. See docs.'),
        ),
        migrations.AlterField(
            model_name='form',
            name='ui_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, help_text='A schema for the form inputs. See docs.', null=True),
        ),
    ]



"""
Use this file to configure pluggable app settings and resolve defaults
with any overrides set in project settings.
"""

from django.conf import settings as project_settings


class Settings:
    pass


Settings.SECRET_KEY = getattr(
    project_settings, "SLACKFORMS_SECRET_KEY", "a-bad-secret-key"
)

Settings.SLACK_VERIFICATION_TOKEN = getattr(
    project_settings, "SLACKFORMS_SLACK_VERIFICATION_TOKEN", ""
)

Settings.SLACK_BOT_TOKEN = getattr(
    project_settings, "SLACKFORMS_SLACK_BOT_TOKEN", ""
)

Settings.SLACK_API_TOKEN = getattr(
    project_settings, "SLACKFORMS_SLACK_API_TOKEN", ""
)

Settings.ROOT_URL = getattr(project_settings, "SLACKFORMS_ROOT_URL", "")

settings = Settings



import dateparser
import re
import requests
import json
import os
import logging
from django.db import models
from django.contrib.postgres.fields import JSONField
from jsonschema import validate, exceptions
from ..utils import schema_to_form, slack, is_float
from ..conf import Settings

logger = logging.getLogger(__name__)


class Form(models.Model):
    """
    A form to be displayed on Slack and will send a validated input to a
    specific endpoint.
    """

    name = models.CharField(
        max_length=20,
        unique=True,
        help_text="The unique name of this form to serve as a unique key.",
    )

    endpoint = models.URLField(
        null=True,
        blank=True,
        help_text="A URL to hit with the processed and validated form data.",
    )

    token = models.ForeignKey(
        "slackforms.Token",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="A token to send with the endpoint.",
    )

    json_schema = JSONField(
        help_text="A schema for the form data. See docs."
    )
    ui_schema = JSONField(
        null=True,
        blank=True,
        help_text="A schema for the form inputs. See docs."
    )
    data_source = models.URLField(
        blank=True,
        help_text="""
The source of data if an Id is supplied. Variable is passed through to
this template. Use "{id}" to crate the template.
""",
    )

    slash_command = models.CharField(
        max_length=40,
        unique=True,
        blank=True,
        null=True,
        help_text="The slash command for this form (don't include the slash).",
    )

    def __str__(self):
        return self.name

    def get_resonse_url(self):
        """
        Get the proper response_url callback for webhook request data.
        """
        return os.path.join(Settings.ROOT_URL, "callback/")

    def get_prop_attr(self, prop, attr):
        """
        Get a particular attribute of a particular form property.
        """
        return (
            self.json_schema.get("properties", {}).get(prop, {}).get(attr, "")
        )

    def is_prop_number(self, prop):
        """
        Get whether a property is of a number type.
        """
        return self.get_prop_attr(prop, "type") in ["number", "integer"]

    def post_ephemeral(self, text, meta):
        username = meta.get("user", {}).get("id", "")
        channel = meta.get("channel", {}).get("id", "")
        slack("chat.postEphemeral", user=username, channel=channel, text=text)

    def get_form_data(self, id, meta):
        """
        Get starting data from the form's data_source given a particular id.
        """
        if not self.data_source == "" and not id == "":
            url = self.data_source.format(id=id)
            r = requests.get(url=url)
            try:
                return r.json()
            except json.decoder.JSONDecodeError:
                self.post_ephemeral(r.text, meta)
                return False
        else:
            return {}

    def process(self, content):
        """
        Process data from a Slack form.
        """
        output = {}
        for prop, value in content.items():
            if value is not None:
                if self.is_prop_number(prop) and is_float(value):
                    output[prop] = int(value)
                elif isinstance(value, float):
                    output[prop] = float(value)
                elif self.get_prop_attr(prop, "type") == "boolean":
                    output[prop] = bool(value)
                elif self.get_prop_attr(prop, "format") == "date-time":
                    try:
                        date = dateparser.parse(value).isoformat()
                    except AttributeError:
                        date = ""
                    output[prop] = date
                else:
                    output[prop] = value

        return output

    def validate(self, content):
        """
        Validate form data based on a form's json schema and return errors
        in a Slack-friendly structure.
        """
        try:
            validate(content, self.json_schema)
        except exceptions.ValidationError as e:
            errorObj = {"name": "".join(e.path), "error": e.message}
            if errorObj["name"] == "":
                m = re.search("'(.*)' is a required property", e.message)
                errorObj["name"] = m.group(1)
            return errorObj

        return True

    def create_slack_form(self, data, data_id):
        """
        Create a Slack form dictionary from the record's json and ui schemas.
        """
        form = schema_to_form(
            self.name, self.json_schema, self.ui_schema, data
        )
        if data_id and not data_id == "":
            form["state"] = data_id

        return form

    def trigger(self, trigger_id, method="POST", **kwargs):
        """
        Trigger a form action. Either a new form or a DELETE request to the
        form's webhook.
        """
        if method == "POST" or method == "PUT":
            self.post_to_slack(trigger_id, **kwargs)
        if method == "DELETE":
            self.post_to_webhook({}, method, **kwargs)

    def post_to_slack(
        self, trigger_id, data_id="", data={}, meta={}, **kwargs
    ):
        """
        Given starting data and/or an Id to get data from a data_source,
        trigger a new form to open in Slack.
        """
        source_data = self.get_form_data(data_id, meta)

        if source_data is False:
            return

        form_data = {**source_data, **data}  # noqa: E999
        form = self.create_slack_form(form_data, data_id=data_id)
        resp = slack("dialog.open", dialog=form, trigger_id=trigger_id)

        if resp.get("ok", False) is True:
            logger.info(resp)
        else:
            logger.error(resp)

    def post_to_webhook(self, processed_content, method, meta={}, **kwargs):
        """
        Send processed form data (along with some metadata) to the form's
        designated webhook.
        """
        if self.endpoint is None:
            return

        data = processed_content
        meta_data = meta
        meta_data["response_url"] = self.get_resonse_url()
        meta_data["form"] = self.name
        meta_data["token"] = self.token.token if self.token is not None else None
        data["slackform_meta_data"] = json.dumps(meta_data)

        if method == "PUT":
            return requests.put(url=self.endpoint, data=data)
        elif method == "DELETE":
            return requests.delete(url=self.endpoint, data=data)
        else:
            return requests.post(url=self.endpoint, data=data)



# Generated by Django 2.1.4 on 2018-12-04 17:28

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0002_auto_20181203_1901'),
    ]

    operations = [
        migrations.AddField(
            model_name='form',
            name='argument_name',
            field=models.CharField(blank=True, help_text='The prop name to map the argument passed from the slash command', max_length=40, null=True, unique=True),
        ),
        migrations.AddField(
            model_name='form',
            name='slash_command',
            field=models.CharField(blank=True, help_text='A slash command mapped to this form', max_length=40, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='form',
            name='json_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(help_text='A schema for the form data. See docs.', null=True),
        ),
        migrations.AlterField(
            model_name='form',
            name='name',
            field=models.CharField(help_text='The unique name of this form to serve as a unique key.', max_length=20, unique=True),
        ),
        migrations.AlterField(
            model_name='form',
            name='ui_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(help_text='A schema for the form inputs. See docs.', null=True),
        ),
        migrations.AlterField(
            model_name='form',
            name='webhook',
            field=models.URLField(blank=True, help_text='A URL to hit with the processed and validated form data'),
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-13 19:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0010_auto_20181213_1923'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='endpoint',
            field=models.ForeignKey(blank=True, help_text='A URL to hit with the processed and validated form data.', null=True, on_delete=django.db.models.deletion.PROTECT, to='slackforms.Endpoint'),
        ),
    ]



from django.http import HttpResponse
from django.core.exceptions import ObjectDoesNotExist
from slackforms.models import Form


class ManualHandler:
    """
    Handler for manual form triggers. Finds the right form from the form
    property in the request data and triggers the appropriate form using the
    trigger_id in the request data.
    """

    def handle(self, data):
        try:
            form = Form.objects.get(name=data.get("form"))
        except ObjectDoesNotExist:
            return HttpResponse(
                "Form {} not found".format(data.get("form")), status=404
            )

        kwargs = {}
        if "data" in data:
            kwargs["data"] = data.get("data")
        if "data_id" in data:
            kwargs["data_id"] = data.get("data_id")

        method = "POST"
        if "method" in data:
            method = data.get("method")

        form.trigger(data.get("trigger_id"), method, **kwargs)

        return HttpResponse("OK", status=200)



from django.http import HttpResponse
from django.core.exceptions import ObjectDoesNotExist
from slackforms.models import Form
from slackforms.utils import slack


class SlashHandler:
    """
    Handler for slash commands. Finds the right form from the name of the
    command being called in the request data and triggers it.
    """

    def handle(self, data):
        try:
            form = Form.objects.get(slash_command=data.get("command")[1:])
        except ObjectDoesNotExist:
            slack(
                "chat.postEphemeral",
                user=data.get("user_id", ""),
                channel=data.get("channel_id", ""),
                text="Form not found for command: `{}`.".format(
                    data.get("command")
                ),
            )
            return HttpResponse(
                "Form {} not found".format(data.get("command")), status=404
            )

        meta = {
            "data_id": data.get("text"),
            "team": {"id": data.get("team_id")},
            "channel": {"id": data.get("channel_id")},
            "user": {"id": data.get("user_id")},
        }

        text = data.get("text")
        method = "POST" if text == "" else "PUT"

        form.trigger(
            data.get("trigger_id"), method=method, data_id=text, meta=meta
        )

        return HttpResponse(status=200)



from django.http import HttpResponse
from django.core.exceptions import ObjectDoesNotExist
from slackforms.models import Form


class InteractionBase:
    """
    Base class for interactive message triggered endpoints. Finds the right
    form from the callback_id and triggers it. Different interactions have
    different locations for their arguments which is coded in child classes.
    """

    def get_id(self):
        return ""

    def get_method(self):
        if self.get_id() == "":
            return "POST"
        else:
            return "PUT"

    def handle(self, data):
        self.data = data

        try:
            self.form = Form.objects.get(name=data.get("callback_id"))
        except ObjectDoesNotExist:
            return HttpResponse(
                "Form {} not found".format(data.get("callback_id")), status=404
            )

        meta = {
            "data_id": self.get_id(),
            "team": data.get("team"),
            "channel": data.get("channel"),
            "user": data.get("user"),
        }

        self.form.trigger(
            data.get("trigger_id"),
            method=self.get_method(),
            data_id=self.get_id(),
            meta=meta,
        )

        return HttpResponse(status=200)



import json
from .base import InteractionBase


class MessageHandler(InteractionBase):
    """
    Handler for buttons and dropdown menus. Both come in using the same request
    type, but their arguments are found in different places. Button values are
    found in the "name" of the button while dropdown menus are found in the
    "value" of the first selected option.
    """

    def menuOrButton(self, actions):
        if "selected_options" in actions[0]:
            return "menu"
        elif "name" in actions[0]:
            return "button"
        else:
            return None

    def get_method(self):
        default = super().get_method()
        actions = self.data.get("actions", [{}])

        if len(actions) == 0:
            return default

        value = {}
        if self.menuOrButton(actions) == "menu":
            value = json.loads(actions[0]["selected_options"][0]["value"])
        elif self.menuOrButton(actions) == "button":
            value = json.loads(actions[0]["value"])

        return value.get("method", default)

    def get_id(self):
        default = super().get_id()
        actions = self.data.get("actions", [{}])

        if len(actions) == 0:
            return default

        value = {}
        if self.menuOrButton(actions) == "menu":
            value = json.loads(actions[0]["selected_options"][0]["value"])
        elif self.menuOrButton(actions) == "button":
            value = json.loads(actions[0]["value"])

        return value.get("data_id", default)



# Generated by Django 2.1.4 on 2018-12-05 16:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0006_auto_20181204_1731'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='form',
            name='argument_prop_name',
        ),
        migrations.AddField(
            model_name='form',
            name='data_source',
            field=models.CharField(blank=True, help_text='\nThe source of data if an Id is supplied. {id} variable passed through. See docs.\n', max_length=40, null=True, unique=True),
        ),
    ]



from .base import InteractionBase


class ActionHandler(InteractionBase):
    """
    Handler for message actions. The argument prop is found in the data's
    "text" property.
    """

    def get_id(self):
        return self.data.get("message", {}).get("text", "")



# Generated by Django 2.1.4 on 2018-12-03 19:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slackforms', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='name',
            field=models.CharField(max_length=20, unique=True),
        ),
    ]
